@@ -1,3 +1,4 @@
+#!/usr/bin/env python
 #
 # SNMP Snapshot Data Recorder
 #
@@ -8,6 +9,8 @@ import getopt
 import time
 import sys
 import os
+import socket
+import traceback
 from pyasn1.type import univ
 from pysnmp.proto import rfc1902, rfc1905
 from pysnmp.entity import engine, config
@@ -24,10 +27,9 @@ from pysnmp.entity.rfc3413 import cmdgen
 from pysnmp import debug
 from snmpsim import __version__
 from snmpsim.record import snmprec
-from snmpsim import confdir, error
+from snmpsim import confdir, error, log
 
 # Defaults
-quietFlag = False
 getBulkFlag = False
 getBulkRepetitions = 25
 snmpVersion = 1
@@ -45,7 +47,6 @@ agentUNIXEndpoint = None
 startOID = univ.ObjectIdentifier('1.3.6')
 stopOID = None
 outputFile = sys.stderr
-variationModulesDirs = []
 variationModuleOptions = ""
 variationModuleName = variationModule = None
 
@@ -65,30 +66,36 @@ privProtocols = {
   'NONE': config.usmNoPrivProtocol
 }
 
-helpMessage = 'Usage: %s [--help] [--debug=<category>] [--quiet] [--version=<1|2c|3>] [--community=<string>] [--v3-user=<username>] [--v3-auth-key=<key>] [--v3-priv-key=<key>] [--v3-auth-proto=<%s>] [--v3-priv-proto=<%s>] [--context=<string>] [--use-getbulk] [--getbulk-repetitions=<number>] [--agent-udpv4-endpoint=<X.X.X.X:NNNNN>] [--agent-udpv6-endpoint=<[X:X:..X]:NNNNN>] [--agent-unix-endpoint=</path/to/named/pipe>] [--start-oid=<OID>] [--stop-oid=<OID>] [--output-file=<filename>] [--variation-modules-dir=<dir>] [--variation-module=<module>] [--variation-module-options=<args]>]' % (sys.argv[0], '|'.join([ x for x in authProtocols if x != 'NONE' ]), '|'.join([ x for x in privProtocols if x != 'NONE' ]))
+helpMessage = 'Usage: %s [--help] [--debug=<category>] [--logging-method=<method|?>] [--version=<1|2c|3>] [--community=<string>] [--v3-user=<username>] [--v3-auth-key=<key>] [--v3-priv-key=<key>] [--v3-auth-proto=<%s>] [--v3-priv-proto=<%s>] [--context=<string>] [--use-getbulk] [--getbulk-repetitions=<number>] [--agent-udpv4-endpoint=<X.X.X.X:NNNNN>] [--agent-udpv6-endpoint=<[X:X:..X]:NNNNN>] [--agent-unix-endpoint=</path/to/named/pipe>] [--start-oid=<OID>] [--stop-oid=<OID>] [--output-file=<filename>] [--variation-modules-dir=<dir>] [--variation-module=<module>] [--variation-module-options=<args>]' % (sys.argv[0], '|'.join([ x for x in authProtocols if x != 'NONE' ]), '|'.join([ x for x in privProtocols if x != 'NONE' ]))
+
+log.setLogger('snmprec', 'stdout')
 
 try:
     opts, params = getopt.getopt(sys.argv[1:], 'h',
-        ['help', 'debug=', 'quiet', 'v1', 'v2c', 'v3', 'version=', 'community=', 'v3-user=', 'v3-auth-key=', 'v3-priv-key=', 'v3-auth-proto=', 'v3-priv-proto=', 'context=', 'use-getbulk', 'getbulk-repetitions=', 'agent-address=', 'agent-port=', 'agent-udpv4-endpoint=', 'agent-udpv6-endpoint=', 'agent-unix-endpoint=', 'start-oid=', 'stop-oid=', 'output-file=', 'variation-modules-dir=', 'variation-module=', 'variation-module-options=']
+        ['help', 'debug=', 'logging-method=', 'quiet', 'v1', 'v2c', 'v3', 'version=', 'community=', 'v3-user=', 'v3-auth-key=', 'v3-priv-key=', 'v3-auth-proto=', 'v3-priv-proto=', 'context=', 'use-getbulk', 'getbulk-repetitions=', 'agent-address=', 'agent-port=', 'agent-udpv4-endpoint=', 'agent-udpv6-endpoint=', 'agent-unix-endpoint=', 'start-oid=', 'stop-oid=', 'output-file=', 'variation-modules-dir=', 'variation-module=', 'variation-module-options=']
         )
 except Exception:
-    sys.stdout.write('getopt error: %s\r\n' % sys.exc_info()[1])
-    sys.stdout.write(helpMessage + '\r\n')
+    sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
     sys.exit(-1)
 
 if params:
-    sys.stdout.write('extra arguments supplied %s\r\n' % params)
-    sys.stdout.write(helpMessage + '\r\n')
+    sys.stderr.write('ERROR: extra arguments supplied %s\r\n%s\r\n' % (params, helpMessage))
     sys.exit(-1)
 
 for opt in opts:
     if opt[0] == '-h' or opt[0] == '--help':
-        sys.stdout.write('SNMP Simulator version %s, written by Ilya Etingof <ilya@glas.net>\r\nSoftware documentation and support at http://snmpsim.sf.net\r\n%s\r\n' % (__version__, helpMessage))
+        sys.stderr.write('SNMP Simulator version %s, written by Ilya Etingof <ilya@glas.net>\r\nSoftware documentation and support at http://snmpsim.sf.net\r\n%s\r\n' % (__version__, helpMessage))
         sys.exit(-1)
     elif opt[0] == '--debug':
         debug.setLogger(debug.Debug(opt[1]))
+    elif opt[0] == '--logging-method':
+        try:
+            log.setLogger('snmprec', *opt[1].split(':'))
+        except error.SnmpsimError:
+            sys.stderr.write('%s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
+            sys.exit(-1)
     elif opt[0] == '--quiet':
-        quietFlag = True
+       log.setLogger('snmprec', 'null') 
     elif opt[0] == '--v1':
         snmpVersion = 0
     elif opt[0] == '--v2c':
@@ -103,7 +110,7 @@ for opt in opts:
         elif opt[1] in ('3', 'v3'):
             snmpVersion = 3
         else:
-            sys.stdout.write('unknown SNMP version %s\r\n' % opt[1])
+            sys.stderr.write('ERROR: unknown SNMP version %s\r\n%s\r\n' % (opt[1], helpMessage))
             sys.exit(-1)
     elif opt[0] == '--community':
         snmpCommunity = opt[1]
@@ -114,14 +121,14 @@ for opt in opts:
     elif opt[0] == '--v3-auth-proto':
         v3AuthProto = opt[1].upper()
         if v3AuthProto not in authProtocols:
-            sys.stdout.write('bad v3 auth protocol %s\r\n' % v3AuthProto)
+            sys.stderr.write('ERROR: bad v3 auth protocol %s\r\n%s\r\n' % (v3AuthProto, helpMessage))
             sys.exit(-1)
     elif opt[0] == '--v3-priv-key':
         v3PrivKey = opt[1]
     elif opt[0] == '--v3-priv-proto':
         v3PrivProto = opt[1].upper()
         if v3PrivProto not in privProtocols:
-            sys.stdout.write('bad v3 privacy protocol %s\r\n' % v3PrivProto)
+            sys.stderr.write('ERROR: bad v3 privacy protocol %s\r\n%s\r\n' % (v3PrivProto, helpMessage))
             sys.exit(-1)
     elif opt[0] == '--context':
         v3Context = opt[1]
@@ -138,26 +145,36 @@ for opt in opts:
         try:
             agentUDPv4Endpoint = f(*opt[1].split(':'))
         except:
-            sys.stdout.write('improper IPv4/UDP endpoint %s\r\n' % opt[1])
+            sys.stderr.write('ERROR: improper IPv4/UDP endpoint %s\r\n%s\r\n' % (opt[1], helpMessage))
+            sys.exit(-1)
+        try:
+            agentUDPv4Endpoint = socket.getaddrinfo(agentUDPv4Endpoint[0], agentUDPv4Endpoint[1], socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)[0][4][:2]
+        except socket.gaierror:
+            sys.stderr.write('ERROR: unknown hostname %s\r\n%s\r\n' % (agentUDPv4Endpoint[0], helpMessage))
             sys.exit(-1)
     elif opt[0] == '--agent-udpv6-endpoint':
         if not udp6:
-            sys.stdout.write('This system does not support UDP/IP6\r\n')
+            sys.stderr.write('This system does not support UDP/IP6\r\n')
             sys.exit(-1)
         if opt[1].find(']:') != -1 and opt[1][0] == '[':
             h, p = opt[1].split(']:')
             try:
                 agentUDPv6Endpoint = h[1:], int(p)
             except:
-                sys.stdout.write('improper IPv6/UDP endpoint %s\r\n' % opt[1])
+                sys.stderr.write('ERROR: improper IPv6/UDP endpoint %s\r\n%s\r\n' % (opt[1], helpMessage))
                 sys.exit(-1)
         elif opt[1][0] == '[' and opt[1][-1] == ']':
             agentUDPv6Endpoint = opt[1][1:-1], 161
         else:
             agentUDPv6Endpoint = opt[1], 161
+        try:
+            agentUDPv6Endpoint = socket.getaddrinfo(agentUDPv6Endpoint[0], agentUDPv6Endpoint[1], socket.AF_INET6, socket.SOCK_DGRAM, socket.IPPROTO_UDP)[0][4][:2]
+        except socket.gaierror:
+            sys.stderr.write('ERROR: unknown hostname %s\r\n%s\r\n' % (agentUDPv6Endpoint[0], helpMessage))
+            sys.exit(-1)
     elif opt[0] == '--agent-unix-endpoint':
         if not unix:
-            sys.stdout.write('This system does not support UNIX domain sockets\r\n')
+            sys.stderr.write('This system does not support UNIX domain sockets\r\n')
             sys.exit(-1)
         agentUNIXEndpoint = opt[1]
     elif opt[0] == '--start-oid':
@@ -167,7 +184,7 @@ for opt in opts:
     elif opt[0] == '--output-file':
         outputFile = open(opt[1], 'wb')
     elif opt[0] == '--variation-modules-dir':
-        variationModulesDirs.append(opt[1])
+        confdir.variation.insert(0, opt[1])
     elif opt[0] == '--variation-module':
         variationModuleName = opt[1]
     elif opt[0] == '--variation-module-options':
@@ -177,74 +194,77 @@ for opt in opts:
 
 if not agentUDPv4Endpoint and not agentUDPv6Endpoint and not agentUNIXEndpoint:
     if agentUDPv4Address[0] is None:
-        sys.stdout.write('ERROR: agent address endpoint not given\r\n%s\r\n' % helpMessage)
+        sys.stderr.write('ERROR: agent endpoint address not specified\r\n%s\r\n' % helpMessage)
         sys.exit(-1)
     else:
         agentUDPv4Endpoint = agentUDPv4Address
 
-if getBulkFlag and not snmpVersion:
-    sys.stdout.write('WARNING: will be using GETNEXT with SNMPv1!\r\n')
-    getBulkFlag = False
-
-# Load variation module
-
-if not variationModulesDirs:
-    [ variationModulesDirs.append(x) for x in confdir.variation ]
-
-for variationModulesDir in variationModulesDirs:
-    sys.stdout.write(
-        'Scanning "%s" directory for variation modules... '%variationModulesDir
-    )
-    if not os.path.exists(variationModulesDir):
-        sys.stdout.write(' no directory\r\n')
-        continue
-
-    if not variationModuleName:
-        sys.stdout.write(' none requested\r\n')
-        break
-
-    mod = os.path.join(variationModulesDir, variationModuleName + '.py')
-
-    ctx = { 'path': mod }
-
-    try:
-        if sys.version_info[0] > 2:
-            exec(compile(open(mod).read(), mod, 'exec'), ctx)
-        else:
-            execfile(mod, ctx)
-    except Exception:
-        sys.stdout.write('\r\nvariation module %s execution failure: %s\r\n' %  (mod, sys.exc_info()[1]))
-        sys.exit(-1)
-    else:
-        variationModule = ctx
-        sys.stdout.write('%s module loaded\r\n' % variationModuleName)
-        break
-else:
-    sys.stdout.write('ERROR: variation module %s not found\r\n%s\r\n' % (variationModuleName, helpMessage))
-    sys.exit(-1)
-
 if snmpVersion == 3:
     if v3User is None:
-        sys.stdout.write('--v3-user is missing\r\n%s\r\n' % helpMessage)
+        sys.stderr.write('ERROR: --v3-user is missing\r\n%s\r\n' % helpMessage)
         sys.exit(-1)
     if v3PrivKey and not v3AuthKey:
-        sys.stdout.write('--v3-auth-key is missing\r\n%s\r\n' % helpMessage)
+        sys.stderr.write('ERROR: --v3-auth-key is missing\r\n%s\r\n' % helpMessage)
         sys.exit(-1)
     if authProtocols[v3AuthProto] == config.usmNoAuthProtocol:
         if v3AuthKey is not None:
             v3AuthProto = 'MD5'
     else:
         if v3AuthKey is None:
-            sys.stdout.write('--v3-auth-key is missing\r\n%s\r\n' % helpMessage)
+            sys.stderr.write('ERROR: --v3-auth-key is missing\r\n%s\r\n' % helpMessage)
             sys.exit(-1)
     if privProtocols[v3PrivProto] == config.usmNoPrivProtocol:
         if v3PrivKey is not None:
             v3PrivProto = 'DES'
     else:
         if v3PrivKey is None:
-            sys.stdout.write('--v3-priv-key is missing\r\n%s\r\n' % helpMessage)
+            sys.stderr.write('ERROR: --v3-priv-key is missing\r\n%s\r\n' % helpMessage)
             sys.exit(-1)
-        
+ 
+if getBulkFlag and not snmpVersion:
+    log.msg('WARNING: will be using GETNEXT with SNMPv1!')
+    getBulkFlag = False
+
+# Attempt to reopen std output stream in binary mode
+if outputFile is sys.stderr:
+    if sys.version_info[0] > 2:
+        outputFile = outputFile.buffer
+    elif sys.platform == "win32":
+        import msvcrt
+        msvcrt.setmode(outputFile.fileno(), os.O_BINARY)
+
+# Load variation module
+
+if variationModuleName:
+    for variationModulesDir in confdir.variation:
+        log.msg('Scanning "%s" directory for variation modules...' % variationModulesDir)
+        if not os.path.exists(variationModulesDir):
+            log.msg('Directory %s does not exist' % variationModulesDir)
+            continue
+
+        mod = os.path.join(variationModulesDir, variationModuleName + '.py')
+        if not os.path.exists(mod):
+            log.msg('Module %s not found' % mod)
+            continue
+
+        ctx = { 'path': mod, 'moduleContext': {} }
+
+        try:
+            if sys.version_info[0] > 2:
+                exec(compile(open(mod).read(), mod, 'exec'), ctx)
+            else:
+                execfile(mod, ctx)
+        except Exception:
+            log.msg('Variation module %s execution failure: %s' %  (mod, sys.exc_info()[1]))
+            sys.exit(-1)
+        else:
+            variationModule = ctx
+            log.msg('Module %s loaded' % variationModuleName)
+            break
+    else:
+        log.msg('ERROR: variation module %s not found' % variationModuleName)
+        sys.exit(-1)
+       
 # SNMP configuration
 
 snmpEngine = engine.SnmpEngine()
@@ -261,16 +281,14 @@ if snmpVersion == 3:
         authProtocols[v3AuthProto], v3AuthKey,
         privProtocols[v3PrivProto], v3PrivKey
         )
-    if not quietFlag:
-        sys.stdout.write('SNMP version 3\r\nContext name: %s\r\nUser: %s\r\nSecurity level: %s\r\nAuthentication key/protocol: %s/%s\r\nEncryption (privacy) key/protocol: %s/%s\r\n' % (v3Context == '' and '\'\'' or v3Context, v3User, secLevel, v3AuthKey is None and '<NONE>' or v3AuthKey, v3AuthProto, v3PrivKey is None and '<NONE>' or v3PrivKey, v3PrivProto))
+    log.msg('SNMP version 3, Context name: %s, SecurityName: %s, SecurityLevel: %s, Authentication key/protocol: %s/%s, Encryption (privacy) key/protocol: %s/%s' % (v3Context == '' and '\'\'' or v3Context, v3User, secLevel, v3AuthKey is None and '<NONE>' or v3AuthKey, v3AuthProto, v3PrivKey is None and '<NONE>' or v3PrivKey, v3PrivProto))
 else:
     v3User = 'agt'
     secLevel = 'noAuthNoPriv'
     config.addV1System(
         snmpEngine, v3User, snmpCommunity
     )
-    if not quietFlag:
-        sys.stdout.write('SNMP version %s\r\nCommunity name: %s\r\n' % (snmpVersion == 0 and '1' or '2c', snmpCommunity))
+    log.msg('SNMP version %s, Community name: %s' % (snmpVersion == 0 and '1' or '2c', snmpCommunity))
 
 config.addTargetParams(snmpEngine, 'pms', v3User, secLevel, snmpVersion)
 
@@ -283,8 +301,7 @@ if agentUDPv6Endpoint:
     config.addTargetAddr(
         snmpEngine, 'tgt', udp6.domainName, agentUDPv6Endpoint, 'pms'
     )
-    if not quietFlag:
-        sys.stdout.write('Querying UDP/IPv6 agent at [%s]:%s\r\n' % agentUDPv6Endpoint)
+    log.msg('Querying UDP/IPv6 agent at [%s]:%s' % agentUDPv6Endpoint)
 elif agentUNIXEndpoint:
     config.addSocketTransport(
         snmpEngine,
@@ -294,8 +311,7 @@ elif agentUNIXEndpoint:
     config.addTargetAddr(
         snmpEngine, 'tgt', unix.domainName, agentUNIXEndpoint, 'pms'
     )
-    if not quietFlag:
-        sys.stdout.write('Querying UNIX named pipe agent at %s\r\n' % agentUNIXEndpoint)
+    log.msg('Querying UNIX named pipe agent at %s' % agentUNIXEndpoint)
 elif agentUDPv4Endpoint:
     config.addSocketTransport(
         snmpEngine,
@@ -305,16 +321,15 @@ elif agentUDPv4Endpoint:
     config.addTargetAddr(
         snmpEngine, 'tgt', udp.domainName, agentUDPv4Endpoint, 'pms'
     )
-    if not quietFlag:
-        sys.stdout.write('Querying UDP/IPv4 agent at %s:%s\r\n' % agentUDPv4Endpoint)
+    log.msg('Querying UDP/IPv4 agent at %s:%s' % agentUDPv4Endpoint)
 
 # Variation module initialization
 
 if variationModule:
-    sys.stdout.write('Initializing variation module:\r\n    %s...' % variationModuleName)
+    log.msg('Initializing variation module...')
     for x in ('init', 'record', 'shutdown'):
         if x not in variationModule:
-            sys.stdout.write('error: missing %s handler!\r\n' % x)
+            log.msg('ERROR: missing %s handler at module %s' % (x, variationModuleName))
             sys.exit(-1)
     try:
         variationModule['init'](snmpEngine,
@@ -323,9 +338,9 @@ if variationModule:
                                 startOID=startOID,
                                 stopOID=stopOID)
     except Exception:
-        sys.stdout.write('FAILED: %s\r\n' % sys.exc_info()[1])
+        log.msg('Module %s initialization FAILED: %s' % (variationModuleName, sys.exc_info()[1]))
     else:
-        sys.stdout.write('OK\r\n')
+        log.msg('Module %s initialization OK' % variationModuleName)
 
 # Data file builder
 
@@ -361,12 +376,12 @@ dataFileHandler = SnmprecRecord()
 def cbFun(sendRequestHandle, errorIndication, errorStatus, errorIndex,
           varBindTable, cbCtx):
     if errorIndication and errorIndication != 'oidNotIncreasing':
-        sys.stdout.write('%s\r\n' % errorIndication)
+        log.msg('SNMP Engine error: %s' % errorIndication)
         return
     # SNMPv1 response may contain noSuchName error *and* SNMPv2c exception,
     # so we ignore noSuchName error here
     if errorStatus and errorStatus != 2:
-        sys.stdout.write('%s\r\n' % errorStatus.prettyPrint())
+        log.msg('Remote SNMP error: %s' % errorStatus.prettyPrint())
         return
 
     stopFlag = False
@@ -424,9 +439,8 @@ def cbFun(sendRequestHandle, errorIndication, errorStatus, errorIndex,
 
             cbCtx['count'] += 1
 
-            if not quietFlag:
-                sys.stdout.write('OIDs dumped: %s/%s\r' % (cbCtx['iteration'], cbCtx['count']))
-                sys.stdout.flush()
+            if cbCtx['count'] % 100 == 0:
+                log.msg('OIDs dumped: %s/%s' % (cbCtx['iteration'], cbCtx['count']))
 
     # Next request time
     cbCtx['reqTime'] = time.time()
@@ -456,7 +470,7 @@ else:
         cbFun, cbCtx, contextName=v3Context
     )
 
-sys.stdout.write('Sending initial %s request....\r\n' % (getBulkFlag and 'GETBULK' or 'GETNEXT'))
+log.msg('Sending initial %s request....' % (getBulkFlag and 'GETBULK' or 'GETNEXT'))
 
 t = time.time()
 
@@ -467,21 +481,20 @@ exc_info = None
 try:
     snmpEngine.transportDispatcher.runDispatcher()
 except KeyboardInterrupt:
-    if not quietFlag:
-        sys.stdout.write('Process terminated\r\n')
+    log.msg('Shutting down process...')
 except Exception:
     exc_info = sys.exc_info()
 
 if variationModule:
-    sys.stdout.write('Shutting down variation modules:\r\n    %s...' % variationModuleName)
+    log.msg('Shutting down variation module %s...' % variationModuleName)
     try:
         variationModule['shutdown'](snmpEngine,
                                     options=variationModuleOptions,
                                     mode='recording')
     except Exception:
-        sys.stdout.write('FAILED: %s\r\n' % sys.exc_info()[1])
+        log.msg('Variation module %s shutdown FAILED: %s' % (variationModuleName, sys.exc_info()[1]))
     else:
-        sys.stdout.write('OK\r\n')
+        log.msg('Variation module %s shutdown OK' % variationModuleName)
 
 snmpEngine.transportDispatcher.closeDispatcher()
 
@@ -489,12 +502,8 @@ t = time.time() - t
 
 cbCtx['total'] += cbCtx['count']
 
-if not quietFlag:
-    sys.stdout.write(
-        'OIDs dumped: %s, elapsed: %.2f sec, rate: %.2f OIDs/sec\r\n' % (cbCtx['total'], t, t and cbCtx['count']//t or 0)
-        )
+log.msg('OIDs dumped: %s, elapsed: %.2f sec, rate: %.2f OIDs/sec' % (cbCtx['total'], t, t and cbCtx['count']//t or 0))
 
 if exc_info:
-    e = exc_info[0](exc_info[1])
-    e.__traceback__ = exc_info[2]
-    raise e
+    for line in traceback.format_exception(*exc_info):
+        log.msg(line.replace('\n', ';'))
