@@ -1,3 +1,4 @@
+#!/usr/bin/env python
 #
 # SNMP Agent Simulator
 #
@@ -7,6 +8,7 @@ import os
 import stat
 import sys
 import getopt
+import traceback
 if sys.version_info[0] < 3 and sys.version_info[1] < 5:
     from md5 import md5
 else:
@@ -34,24 +36,24 @@ from pysnmp import error
 from pysnmp import debug
 from snmpsim import __version__
 from snmpsim.error import SnmpsimError, NoDataNotification
-from snmpsim import confdir
+from snmpsim import confdir, log, daemon
 from snmpsim.record import dump, mvc, sap, walk, snmprec
 from snmpsim.record.search.file import searchRecordByOid
 from snmpsim.record.search.database import RecordIndex
 
-# Process command-line options
-
-# Defaults
+# Settings
 forceIndexBuild = False
 validateData = False
 v2cArch = False
 v3Only = False
-v3User = 'simulator'
-v3AuthKey = 'auctoritas'
-v3AuthProto = 'MD5'
-v3PrivKey = 'privatus'
-v3PrivProto = 'DES'
-agentUDPv4Address = ('127.0.0.1', 161)
+v3Users = []
+v3AuthKeys = {}
+v3AuthProtos = {}
+v3PrivKeys = {}
+v3PrivProtos = {}
+pidFile = '/var/run/snmpsim/snmpsimd.pid'
+foregroundFlag = True
+procUser = procGroup = None
 agentUDPv4Endpoints = []
 agentUDPv6Endpoints = []
 agentUNIXEndpoints = []
@@ -74,173 +76,6 @@ privProtocols = {
   'NONE': config.usmNoPrivProtocol
 }
  
-helpMessage = 'Usage: %s [--help] [--version ] [--debug=<category>] [--data-dir=<dir>] [--cache-dir=<dir>] [--force-index-rebuild] [--validate-data] [--variation-modules-dir=<dir>] [--variation-module-options=<module[=alias][:args]>] [--agent-udpv4-endpoint=<X.X.X.X:NNNNN>] [--agent-udpv6-endpoint=<[X:X:..X]:NNNNN>] [--agent-unix-endpoint=</path/to/named/pipe>] [--v2c-arch] [--v3-only] [--v3-user=<username>] [--v3-auth-key=<key>] [--v3-auth-proto=<%s>] [--v3-priv-key=<key>] [--v3-priv-proto=<%s>]' % (sys.argv[0], '|'.join(authProtocols), '|'.join(privProtocols))
-
-try:
-    opts, params = getopt.getopt(sys.argv[1:], 'h',
-        ['help', 'debug=', 'device-dir=', 'data-dir=', 'cache-dir=', 'force-index-rebuild', 'validate-device-data', 'validate-data', 'variation-modules-dir=', 'variation-module-options=', 'agent-address=', 'agent-port=', 'agent-udpv4-endpoint=', 'agent-udpv6-endpoint=', 'agent-unix-endpoint=', 'v2c-arch', 'v3-only', 'v3-user=', 'v3-auth-key=', 'v3-auth-proto=', 'v3-priv-key=', 'v3-priv-proto=']
-        )
-except Exception:
-    sys.stdout.write('%s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
-    sys.exit(-1)
-
-if params:
-    sys.stdout.write('extra arguments supplied %s%s\r\n' % (params, helpMessage))
-    sys.exit(-1)
-
-for opt in opts:
-    if opt[0] == '-h' or opt[0] == '--help' or \
-       opt[0] == '-v' or opt[0] == '--version':
-        sys.stdout.write('SNMP Simulator version %s, written by Ilya Etingof <ilya@glas.net>\r\nSoftware documentation and support at http://snmpsim.sf.net\r\n%s\r\n' % (__version__, helpMessage))
-        sys.exit(-1)
-    elif opt[0] == '--debug':
-        debug.setLogger(debug.Debug(opt[1]))
-    elif opt[0] in ('--device-dir', '--data-dir'):
-        confdir.data.insert(0, opt[1])
-    elif opt[0] == '--cache-dir':
-        confdir.cache = opt[1]
-    elif opt[0] == '--force-index-rebuild':
-        forceIndexBuild = True
-    elif opt[0] in ('--validate-device-data', '--validate-data'):
-        validateData = True
-    elif opt[0] == '--variation-modules-dir':
-        confdir.variation.insert(0, opt[1])
-    elif opt[0] == '--variation-module-options':
-        args = opt[1].split(':', 1)
-        try:
-            modName, args = args[0], args[1]
-        except:
-            sys.stdout.write('improper variation module options: %s\r\n'%opt[1])
-            sys.exit(-1)
-        if '=' in modName:
-            modName, alias = modName.split('=', 1)
-        else:
-            alias = os.path.splitext(os.path.basename(modName))[0]
-        if modName not in variationModulesOptions:
-            variationModulesOptions[modName] = []
-        variationModulesOptions[modName].append((alias, args))
-    elif opt[0] == '--agent-udpv4-endpoint':
-        f = lambda h,p=161: (h, int(p))
-        try:
-            agentUDPv4Endpoints.append(f(*opt[1].split(':')))
-        except:
-            sys.stdout.write('improper IPv4/UDP endpoint %s\r\n' % opt[1])
-            sys.exit(-1)
-    elif opt[0] == '--agent-udpv6-endpoint':
-        if not udp6:
-            sys.stdout.write('This system does not support UDP/IP6\r\n')
-            sys.exit(-1)
-        if opt[1].find(']:') != -1 and opt[1][0] == '[':
-            h, p = opt[1].split(']:')
-            try:
-                h, p = h[1:], int(p)
-            except:
-                sys.stdout.write('improper IPv6/UDP endpoint %s\r\n' % opt[1])
-                sys.exit(-1)
-        elif opt[1][0] == '[' and opt[1][-1] == ']':
-            h, p = opt[1][1:-1], 161
-        else:
-            h, p = opt[1], 161
-        agentUDPv6Endpoints.append((h, p))
-    elif opt[0] == '--agent-unix-endpoint':
-        if not unix:
-            sys.stdout.write('This system does not support UNIX domain sockets\r\n')
-            sys.exit(-1)
-        agentUNIXEndpoints.append(opt[1])
-    elif opt[0] == '--agent-address':
-        agentUDPv4Address = (opt[1], agentUDPv4Address[1])
-    elif opt[0] == '--agent-port':
-        agentUDPv4Address = (agentUDPv4Address[0], int(opt[1]))
-    elif opt[0] == '--v2c-arch':
-        v2cArch = True
-    elif opt[0] == '--v3-only':
-        v3Only = True
-    elif opt[0] == '--v3-user':
-        v3User = opt[1]
-    elif opt[0] == '--v3-auth-key':
-        v3AuthKey = opt[1]
-    elif opt[0] == '--v3-auth-proto':
-        v3AuthProto = opt[1].upper()
-        if v3AuthProto not in authProtocols:
-            sys.stdout.write('bad v3 auth protocol %s\r\n' % v3AuthProto)
-            sys.exit(-1)
-    elif opt[0] == '--v3-priv-key':
-        v3PrivKey = opt[1]
-    elif opt[0] == '--v3-priv-proto':
-        v3PrivProto = opt[1].upper()
-        if v3PrivProto not in privProtocols:
-            sys.stdout.write('bad v3 privacy protocol %s\r\n' % v3PrivProto)
-            sys.exit(-1)
-
-if authProtocols[v3AuthProto] == config.usmNoAuthProtocol and \
-    privProtocols[v3PrivProto] != config.usmNoPrivProtocol:
-        sys.stdout.write('privacy impossible without authentication\r\n')
-        sys.exit(-1)
-
-for variationModulesDir in confdir.variation:
-    sys.stdout.write(
-        'Scanning "%s" directory for variation modules... '%variationModulesDir
-    )
-    if not os.path.exists(variationModulesDir):
-        sys.stdout.write(' no directory\r\n')
-        continue
-    for dFile in os.listdir(variationModulesDir):
-        if dFile[-3:] != '.py':
-            continue
-        _toLoad = []
-        modName = os.path.splitext(os.path.basename(dFile))[0]
-        if modName in variationModulesOptions:
-            while variationModulesOptions[modName]:
-                alias, args = variationModulesOptions[modName].pop()
-                _toLoad.append((alias, args))
-            del variationModulesOptions[modName]
-        else:
-            _toLoad.append((modName, ''))
-
-        mod = variationModulesDir + os.path.sep + dFile
-
-        for alias, args in _toLoad:
-            if alias in variationModules:
-                sys.stdout.write('\r\nWARNING: ignoring duplicate module %s at %s\r\n' %  (alias, mod))
-                continue
-
-            ctx = { 'path': mod,
-                    'alias': alias,
-                    'args': args }
-
-            try:
-                if sys.version_info[0] > 2:
-                    exec(compile(open(mod).read(), mod, 'exec'), ctx)
-                else:
-                    execfile(mod, ctx)
-            except Exception:
-                sys.stdout.write('\r\nvariation module %s execution failure: %s\r\n' %  (mod, sys.exc_info()[1]))
-                sys.exit(-1)
-            else:
-                variationModules[alias] = ctx
-
-    sys.stdout.write('%s more modules found\r\n' % len(variationModules))
-
-if variationModulesOptions:
-    sys.stdout.write('ERROR: unused options for variation modules: %s\r\n' %  ', '.join(variationModulesOptions.keys()))
-    sys.exit(-1)
-     
-# for backward compatibility
-if not agentUDPv4Endpoints and \
-   not agentUDPv6Endpoints and \
-   not agentUNIXEndpoints:
-    agentUDPv4Endpoints.append(agentUDPv4Address)
-
-if not os.path.exists(confdir.cache):
-    sys.stdout.write('Creating cache directory %s... \r' % confdir.cache)
-    try:
-        os.makedirs(confdir.cache)
-    except OSError:
-        sys.stdout.write('ERROR: %s: %s\r\n' % (confdir.cache, sys.exc_info()[1]))
-        sys.exit(-1)
-    else:
-        sys.stdout.write('done\r\n')
-
 # Extended snmprec record handler
 
 class SnmprecRecord(snmprec.SnmprecRecord):
@@ -262,10 +97,22 @@ class SnmprecRecord(snmprec.SnmprecRecord):
                         if hexvalue is not None:
                             context['hexvalue'] = hexvalue
                             context['hextag'] = self.grammar.getTagByType(context['origValue']) + 'x'
+
+                    # prepare agent and record contexts on first reference
+                    ( variationModule,
+                      agentContexts,
+                      recordContexts ) = context['variationModules'][modName]
+                    if context['dataFile'] not in agentContexts:
+                        agentContexts[context['dataFile']] = {}
+                    variationModule['agentContext'] = agentContexts[context['dataFile']]
+                    if oid not in recordContexts:
+                        recordContexts[oid] = {}
+                    variationModule['recordContext'] = recordContexts[oid]
+
                     # invoke variation module
-                    oid, tag, value = context['variationModules'][modName]['variate'](oid, tag, value, **context)
+                    oid, tag, value = variationModule['variate'](oid, tag, value, **context)
             else:
-                raise SnmpsimError('variation module "%s" referenced but not loaded\r\n' % modName)
+                raise SnmpsimError('Variation module "%s" referenced but not loaded\r\n' % modName)
 
         if not modName:
             if 'dataValidation' in context:
@@ -332,11 +179,14 @@ class DataFile(AbstractLayout):
     def getHandles(self):
         if not self.__recordIndex.isOpen():
             if len(DataFile.openedQueue) > self.maxQueueEntries:
+                log.msg('Closing %s' % self)
                 DataFile.openedQueue[0].close()
                 del DataFile.openedQueue[0]
 
             DataFile.openedQueue.append(self)
 
+            log.msg('Opening %s' % self)
+
             self.__recordIndex.open()
 
         return self.__recordIndex.getHandles()
@@ -352,7 +202,9 @@ class DataFile(AbstractLayout):
         text, db = self.getHandles()
        
         varsRemaining = varsTotal = len(varBinds)
-         
+        
+        log.msg('Request var-binds: %s, flags: %s, %s' % (', '.join(['%s=<%s>' % (vb[0], vb[1].prettyPrint()) for vb in varBinds]), nextFlag and 'NEXT' or 'EXACT', setFlag and 'SET' or 'GET'))
+
         for oid, val in varBinds:
             textOid = str(
                 univ.OctetString('.'.join([ '%s' % x for x in oid ]))
@@ -376,6 +228,17 @@ class DataFile(AbstractLayout):
             varsRemaining -= 1
 
             line = text.readline()  # matched line
+           
+            # case of a subtree on the last line 
+            if not exactMatch and not line:
+                offset -= 1
+                while offset > 0:
+                    offset -= 1
+                    text.seek(offset)
+                    if text.read(1) == '\n':
+                        break
+                text.seek(offset)
+                line = text.readline()  # last line
 
             while True:
                 if exactMatch:
@@ -386,7 +249,7 @@ class DataFile(AbstractLayout):
                             try:
                                 _, subtreeFlag, _ = self.__recordIndex.lookup(str(_nextOid)).split(str2octs(','))
                             except KeyError:
-                                sys.stdout.write('data error for %s at %s, index broken?\r\n' % (self, _nextOid))
+                                log.msg('data error for %s at %s, index broken?' % (self, _nextOid))
                                 line = ''  # fatal error
                                 break
                             subtreeFlag = int(subtreeFlag)
@@ -399,7 +262,7 @@ class DataFile(AbstractLayout):
                         try:
                             _, _, _prevOffset = self.__recordIndex.lookup(str(_oid)).split(str2octs(','))
                         except KeyError:
-                            sys.stdout.write('data error for %s at %s, index broken?\r\n' % (self, _oid))
+                            log.msg('data error for %s at %s, index broken?' % (self, _oid))
                             line = ''  # fatal error
                             break
                         _prevOffset = int(_prevOffset)
@@ -433,17 +296,17 @@ class DataFile(AbstractLayout):
                 except Exception:
                     _oid = oid
                     _val = errorStatus
-                    sys.stdout.write(
-                        'data error at %s for %s: %s\r\n' % (self, textOid, sys.exc_info()[1])
-                    )
+                    log.msg('data error at %s for %s: %s' % (self, textOid, sys.exc_info()[1]))
 
                 break
 
             rspVarBinds.append((_oid, _val))
 
+        log.msg('Response var-binds: %s' % (', '.join(['%s=<%s>' % (vb[0], vb[1].prettyPrint()) for vb in rspVarBinds])))
+
         return rspVarBinds
  
-    def __str__(self): return str(self.__recordIndex)
+    def __str__(self): return '%s controller' % self.__textFile
 
 # Collect data files
 
@@ -503,6 +366,8 @@ class DataIndexInstrumController:
         self.__indexOid = baseOid + self.indexSubOid
         self.__idx = 1
 
+    def __str__(self): return '<index> controller'
+
     def readVars(self, varBinds, acInfo=None):
         return [ (vb[0], self.__db.get(vb[0], exval.noSuchInstance)) for vb in varBinds ]
 
@@ -550,23 +415,25 @@ def probeContext(transportDomain, transportAddress, contextName):
     candidate = [ str(x) for x in candidate if x ]
 
     while candidate:
-        yield rfc1902.OctetString(
-                  os.path.normpath(os.path.sep.join(candidate))
-              ).asOctets()
+        yield rfc1902.OctetString(os.path.normpath(os.path.sep.join(candidate)).replace(os.path.sep, '/')).asOctets()
         del candidate[-1]
  
 if not v2cArch:
     def probeHashContext(self, snmpEngine, stateReference, contextName):
         transportDomain, transportAddress = snmpEngine.msgAndPduDsp.getTransportInfo(stateReference)
 
-        for probedContextName in probeContext(transportDomain, transportAddress, contextName):
-            probedContextName = md5(probedContextName).hexdigest()
+        for candidate in probeContext(transportDomain, transportAddress, contextName):
+            probedContextName = md5(candidate).hexdigest()
             try:
-                self.snmpContext.getMibInstrum(probedContextName)
+                mibInstrum = self.snmpContext.getMibInstrum(probedContextName)
             except error.PySnmpError:
                 pass
             else:
+                log.msg('Using %s selected by candidate %s; transport ID %s, source address %s, community name "%s"' % (mibInstrum, candidate, univ.ObjectIdentifier(transportDomain), transportAddress[0], communityName))
                 return probedContextName
+        else:
+            log.msg('Using %s selected by contextName "%s", transport ID %s, source address %s' % (self.snmpContext.getMibInstrum(contextName), contextName, univ.ObjectIdentifier(transportDomain), transportAddress[0]))
+
         return contextName
 
     class GetCommandResponder(cmdrsp.GetCommandResponder):
@@ -609,6 +476,227 @@ if not v2cArch:
                 PDU, acInfo
             )
 
+# main script body starts here
+
+helpMessage = 'Usage: %s [--help] [--version ] [--debug=<category>] [--daemonize] [--process-user=<uname>] [--process-group=<gname>] [--logging-method=<method|?>] [--data-dir=<dir>] [--cache-dir=<dir>] [--force-index-rebuild] [--validate-data] [--variation-modules-dir=<dir>] [--variation-module-options=<module[=alias][:args]>] [--agent-udpv4-endpoint=<X.X.X.X:NNNNN>] [--agent-udpv6-endpoint=<[X:X:..X]:NNNNN>] [--agent-unix-endpoint=</path/to/named/pipe>] [--v2c-arch] [--v3-only] [--v3-user=<username>] [--v3-auth-key=<key>] [--v3-auth-proto=<%s>] [--v3-priv-key=<key>] [--v3-priv-proto=<%s>]' % (sys.argv[0], '|'.join(authProtocols), '|'.join(privProtocols))
+
+try:
+    opts, params = getopt.getopt(sys.argv[1:], 'h',
+        ['help', 'debug=', 'daemonize', 'process-user=', 'process-group=', 'logging-method=', 'device-dir=', 'data-dir=', 'cache-dir=', 'force-index-rebuild', 'validate-device-data', 'validate-data', 'variation-modules-dir=', 'variation-module-options=', 'agent-address=', 'agent-port=', 'agent-udpv4-endpoint=', 'agent-udpv6-endpoint=', 'agent-unix-endpoint=', 'v2c-arch', 'v3-only', 'v3-user=', 'v3-auth-key=', 'v3-auth-proto=', 'v3-priv-key=', 'v3-priv-proto=']
+        )
+except Exception:
+    sys.stderr.write('ERROR: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
+    sys.exit(-1)
+
+if params:
+    sys.stderr.write('ERROR: extra arguments supplied %s\r\n%s\r\n' % (params, helpMessage))
+    sys.exit(-1)
+
+log.setLogger('snmpsimd', 'stdout')
+
+for opt in opts:
+    if opt[0] == '-h' or opt[0] == '--help' or \
+       opt[0] == '-v' or opt[0] == '--version':
+        sys.stderr.write('SNMP Simulator version %s, written by Ilya Etingof <ilya@glas.net>\r\nSoftware documentation and support at http://snmpsim.sf.net\r\n%s\r\n' % (__version__, helpMessage))
+        sys.exit(-1)
+    elif opt[0] == '--debug':
+        debug.setLogger(debug.Debug(opt[1]))
+    elif opt[0] == '--daemonize':
+        foregroundFlag = False
+    elif opt[0] == '--process-user':
+        procUser = opt[1]
+    elif opt[0] == '--process-group':
+        procGroup = opt[1]
+    elif opt[0] == '--logging-method':
+        try:
+            log.setLogger('snmpsimd', *opt[1].split(':'))
+        except SnmpsimError:
+            sys.stderr.write('%s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
+            sys.exit(-1)
+    elif opt[0] in ('--device-dir', '--data-dir'):
+        confdir.data.insert(0, opt[1])
+    elif opt[0] == '--cache-dir':
+        confdir.cache = opt[1]
+    elif opt[0] == '--force-index-rebuild':
+        forceIndexBuild = True
+    elif opt[0] in ('--validate-device-data', '--validate-data'):
+        validateData = True
+    elif opt[0] == '--variation-modules-dir':
+        confdir.variation.insert(0, opt[1])
+    elif opt[0] == '--variation-module-options':
+        args = opt[1].split(':', 1)
+        try:
+            modName, args = args[0], args[1]
+        except:
+            sys.stderr.write('ERROR: improper variation module options: %s\r\n%s\r\n' % (opt[1], helpMessage))
+            sys.exit(-1)
+        if '=' in modName:
+            modName, alias = modName.split('=', 1)
+        else:
+            alias = os.path.splitext(os.path.basename(modName))[0]
+        if modName not in variationModulesOptions:
+            variationModulesOptions[modName] = []
+        variationModulesOptions[modName].append((alias, args))
+    elif opt[0] == '--agent-udpv4-endpoint':
+        f = lambda h,p=161: (h, int(p))
+        try:
+            agentUDPv4Endpoints.append((udp.UdpTransport().openServerMode(f(*opt[1].split(':'))), opt[1]))
+        except:
+            sys.stderr.write('ERROR: improper/busy IPv4/UDP endpoint %s\r\n%s\r\n' % (opt[1], helpMessage))
+            sys.exit(-1)
+    elif opt[0] == '--agent-udpv6-endpoint':
+        if not udp6:
+            sys.stderr.write('This system does not support UDP/IP6\r\n')
+            sys.exit(-1)
+        if opt[1].find(']:') != -1 and opt[1][0] == '[':
+            h, p = opt[1].split(']:')
+            try:
+                h, p = h[1:], int(p)
+            except:
+                sys.stderr.write('ERROR: improper/busy IPv6/UDP endpoint %s\r\n%s\r\n' % (opt[1], helpMessage))
+                sys.exit(-1)
+        elif opt[1][0] == '[' and opt[1][-1] == ']':
+            h, p = opt[1][1:-1], 161
+        else:
+            h, p = opt[1], 161
+        agentUDPv6Endpoints.append(
+            (udp6.Udp6Transport().openServerMode((h, p)), opt[1])
+        )
+    elif opt[0] == '--agent-unix-endpoint':
+        if not unix:
+            sys.stderr.write('This system does not support UNIX domain sockets\r\n')
+            sys.exit(-1)
+        agentUNIXEndpoints.append(
+            (unix.UnixTransport().openServerMode(opt[1]), opt[1])
+        )
+    elif opt[0] == '--agent-address':
+        sys.stderr.write('ERROR: use --agent-udpv4-endpoint=%s option instead of --agent-address\r\n%s\r\n' % (opt[1], helpMessage))
+        sys.exit(-1)
+    elif opt[0] == '--agent-port':
+        sys.stderr.write('ERROR: use --agent-udpv4-endpoint=0.0.0.0:%s option instead of --agent-port\r\n%s\r\n' % (opt[1], helpMessage))
+        sys.exit(-1)
+    elif opt[0] == '--v2c-arch':
+        v2cArch = True
+    elif opt[0] == '--v3-only':
+        v3Only = True
+    elif opt[0] == '--v3-user':
+        v3Users.append(opt[1])
+    elif opt[0] == '--v3-auth-key':
+        v3AuthKeys[v3Users[-1]] = opt[1]
+    elif opt[0] == '--v3-auth-proto':
+        v3AuthProtos[v3Users[-1]] = opt[1].upper()
+        if v3AuthProtos[v3Users[-1]] not in authProtocols:
+            sys.stderr.write('ERROR: bad v3 auth protocol %s\r\n%s\r\n' % (v3AuthProtos[v3Users[-1]], helpMessage))
+            sys.exit(-1)
+    elif opt[0] == '--v3-priv-key':
+        v3PrivKeys[v3Users[-1]] = opt[1]
+    elif opt[0] == '--v3-priv-proto':
+        v3PrivProtos[v3Users[-1]] = opt[1].upper()
+        if v3PrivProtos[v3Users[-1]] not in privProtocols:
+            sys.stderr.write('ERROR: bad v3 privacy protocol %s\r\n%s\r\n' % (v3PrivProtos[v3Users[-1]], helpMessage))
+            sys.exit(-1)
+
+for v3User in v3Users:
+    if v3User in v3AuthKeys:
+        if v3User not in v3AuthProtos:
+            v3AuthProtos[v3User] = 'MD5'
+    else:
+        v3AuthKeys[v3User] = None
+        v3AuthProtos[v3User] = 'NONE'
+    if v3User in v3PrivKeys:
+        if v3User not in v3PrivProtos:
+            v3PrivProtos[v3User] = 'DES'
+    else:
+        v3PrivKeys[v3User] = None
+        v3PrivProtos[v3User] = 'NONE'
+    if authProtocols[v3AuthProtos[v3User]] == config.usmNoAuthProtocol and \
+        privProtocols[v3PrivProtos[v3User]] != config.usmNoPrivProtocol:
+            sys.stderr.write('ERROR: privacy impossible without authentication\r for USM user %s\n%s\r\n', (v3User, helpMessage))
+            sys.exit(-1)
+
+if not v3Users:
+    v3Users = [ 'simulator' ]
+    v3AuthKeys[v3Users[0]] = 'auctoritas'
+    v3AuthProtos[v3Users[0]] = 'MD5'
+    v3PrivKeys[v3Users[0]] = 'privatus'
+    v3PrivProtos[v3Users[0]] = 'DES'
+
+if not agentUDPv4Endpoints and \
+   not agentUDPv6Endpoints and \
+   not agentUNIXEndpoints:
+        sys.stderr.write('ERROR: agent endpoint address(es) not specified\r\n%s\r\n' % helpMessage)
+        sys.exit(-1)
+
+try:
+    daemon.dropPrivileges(procUser, procGroup)
+except:
+    sys.stderr.write('ERROR: cant drop priveleges: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
+    sys.exit(-1)
+
+if not foregroundFlag:
+    try:
+        daemon.daemonize(pidFile)
+    except:
+        sys.stderr.write('ERROR: cant daemonize process: %s\r\n%s\r\n' % (sys.exc_info()[1], helpMessage))
+        sys.exit(-1)
+
+for variationModulesDir in confdir.variation:
+    log.msg('Scanning "%s" directory for variation modules...' % variationModulesDir)
+    if not os.path.exists(variationModulesDir):
+        log.msg('Directory %s does not exist' % variationModulesDir)
+        continue
+    for dFile in os.listdir(variationModulesDir):
+        if dFile[-3:] != '.py':
+            continue
+        _toLoad = []
+        modName = os.path.splitext(os.path.basename(dFile))[0]
+        if modName in variationModulesOptions:
+            while variationModulesOptions[modName]:
+                alias, args = variationModulesOptions[modName].pop()
+                _toLoad.append((alias, args))
+            del variationModulesOptions[modName]
+        else:
+            _toLoad.append((modName, ''))
+
+        mod = variationModulesDir + os.path.sep + dFile
+
+        for alias, args in _toLoad:
+            if alias in variationModules:
+                log.msg('WARNING: ignoring duplicate module %s at %s' %  (alias, mod))
+                continue
+
+            ctx = { 'path': mod,
+                    'alias': alias,
+                    'args': args,
+                    'moduleContext': {} }
+
+            try:
+                if sys.version_info[0] > 2:
+                    exec(compile(open(mod).read(), mod, 'exec'), ctx)
+                else:
+                    execfile(mod, ctx)
+            except Exception:
+                log.msg('Variation module %s execution failure: %s' %  (mod, sys.exc_info()[1]))
+                sys.exit(-1)
+            else:
+                # moduleContext, agentContext, recordContext
+                variationModules[alias] = ctx, {}, {}
+
+    log.msg('A total of %s modules found in %s' % (len(variationModules), variationModulesDir))
+
+if variationModulesOptions:
+    log.msg('WARNING: unused options for variation modules: %s' %  ', '.join(variationModulesOptions.keys()))
+
+if not os.path.exists(confdir.cache):
+    try:
+        os.makedirs(confdir.cache)
+    except OSError:
+        log.msg('ERROR: failed to create cache directory %s: %s' % (confdir.cache, sys.exc_info()[1]))
+        sys.exit(-1)
+    else:
+        log.msg('Cache directory %s created' % confdir.cache)
+
+
 # Basic SNMP engine configuration
 
 if v2cArch:
@@ -620,29 +708,30 @@ else:
 
     snmpContext = context.SnmpContext(snmpEngine)
         
-    config.addV3User(
-        snmpEngine,
-        v3User,
-        authProtocols[v3AuthProto], v3AuthKey,
-        privProtocols[v3PrivProto], v3PrivKey
+    for v3User in v3Users:
+        config.addV3User(
+            snmpEngine,
+            v3User,
+            authProtocols[v3AuthProtos[v3User]], v3AuthKeys[v3User],
+            privProtocols[v3PrivProtos[v3User]], v3PrivKeys[v3User]
         )
 
 if variationModules:
-    sys.stdout.write('Initializing variation modules:\r\n')
-    for name, body in variationModules.items():
-        sys.stdout.write('    %s...  ' % name)
+    log.msg('Initializing variation modules...')
+    for name, contexts in variationModules.items():
+        body = contexts[0]
         for x in ('init', 'variate', 'shutdown'):
             if x not in body:
-                sys.stdout.write('error: missing %s handler!\r\n' % x)
+                log.msg('ERROR: missing %s handler in %s!' % (x, name))
                 sys.exit(-1)
         try:
             body['init'](not v2cArch and snmpEngine or None,
                          options=body['args'],
                          mode='variating')
         except Exception:
-            sys.stdout.write('FAILED: %s\r\n' % sys.exc_info()[1])
+            log.msg('Module %s load FAILED: %s' % (name,sys.exc_info()[1]))
         else:
-            sys.stdout.write('OK\r\n')
+            log.msg('Modile %s loaded OK' % name)
 
 # Build pysnmp Managed Objects base from data files information
 
@@ -650,20 +739,19 @@ _mibInstrums = {}
 _dataFiles = {}
 
 for dataDir in confdir.data:
-    sys.stdout.write(
-        'Scanning "%s" directory for %s data files...' % (dataDir, ','.join([' *%s%s' % (os.path.extsep, x.ext) for x in recordSet.values()]))
+    log.msg('Scanning "%s" directory for %s data files...' % (dataDir, ','.join([' *%s%s' % (os.path.extsep, x.ext) for x in recordSet.values()]))
     )
     if not os.path.exists(dataDir):
-        sys.stdout.write(' no directory\r\n')
+        log.msg('Directory %s does not exist' % dataDir)
         continue
-    sys.stdout.write('\r\n%s\r\n' % ('='*66,))
+    log.msg('%s' % ('='*66,))
     for fullPath, textParser, communityName in getDataFiles(dataDir):
         if communityName in _dataFiles:
-            sys.stdout.write('WARNING: ignoring duplicate Community/ContextName "%s" for data file %s (%s already loaded)\r\n' % (communityName, fullPath, _dataFiles[communityName]))
+            log.msg('WARNING: ignoring duplicate Community/ContextName "%s" for data file %s (%s already loaded)' % (communityName, fullPath, _dataFiles[communityName]))
             continue
         elif fullPath in _mibInstrums:
             mibInstrum = _mibInstrums[fullPath]
-            sys.stdout.write('Shared %s\r\n' % (mibInstrum,))
+            log.msg('Configuring *shared* %s' % (mibInstrum,))
         else:
             dataFile = DataFile(fullPath, textParser).indexText(forceIndexBuild)
             mibInstrum = mibInstrumControllerSet[dataFile.layout](dataFile)
@@ -671,9 +759,9 @@ for dataDir in confdir.data:
             _mibInstrums[fullPath] = mibInstrum
             _dataFiles[communityName] = fullPath
 
-            sys.stdout.write('%s\r\n' % (mibInstrum,))
+            log.msg('Configuring %s' % (mibInstrum,))
 
-        sys.stdout.write('SNMPv1/2c community name: %s\r\n' % (communityName,))
+        log.msg('SNMPv1/2c community name: %s' % (communityName,))
 
         if v2cArch:
             contexts[univ.OctetString(communityName)] = mibInstrum
@@ -695,9 +783,9 @@ for dataDir in confdir.data:
                 fullPath, communityName, contextName
             )
                  
-            sys.stdout.write('SNMPv3 context name: %s\r\n' % (contextName,))
+            log.msg('SNMPv3 context name: %s' % (contextName,))
         
-        sys.stdout.write('%s\r\n' % ('-+' * 33,))
+        log.msg('%s' % ('-+' * 33,))
         
 del _mibInstrums
 del _dataFiles
@@ -729,17 +817,20 @@ if v2cArch:
             if msgVer in api.protoModules:
                 pMod = api.protoModules[msgVer]
             else:
-                sys.stdout.write('Unsupported SNMP version %s\r\n' % (msgVer,))
+                log.msg('Unsupported SNMP version %s' % (msgVer,))
                 return
             reqMsg, wholeMsg = decoder.decode(
                 wholeMsg, asn1Spec=pMod.Message(),
                 )
 
             communityName = reqMsg.getComponentByPosition(1)
-            for communityName in probeContext(transportDomain, transportAddress, communityName):
-                if communityName in contexts:
+            for candidate in probeContext(transportDomain, transportAddress, communityName):
+                if candidate in contexts:
+                    log.msg('Using %s selected by candidate %s; transport ID %s, source address %s, community name "%s"' % (contexts[candidate], candidate, univ.ObjectIdentifier(transportDomain), transportAddress[0], communityName))
+                    communityName = candidate
                     break
             else:
+                log.msg('No data file selected for transport ID %s, source address %s, community name "%s"' % (univ.ObjectIdentifier(transportDomain), transportAddress[0], communityName))
                 return wholeMsg
             
             rspMsg = pMod.apiMessage.getResponse(reqMsg)
@@ -755,7 +846,7 @@ if v2cArch:
             elif hasattr(pMod, 'GetBulkRequestPDU') and \
                      reqPDU.isSameTypeWith(pMod.GetBulkRequestPDU()):
                 if not msgVer:
-                    sys.stdout.write('GETBULK over SNMPv1 from %s:%s\r\n' % (
+                    log.msg('GETBULK over SNMPv1 from %s:%s' % (
                         transportDomain, transportAddress
                         ))
                     return wholeMsg
@@ -764,7 +855,7 @@ if v2cArch:
                     pMod.apiBulkPDU.getMaxRepetitions(reqPDU),
                     contexts[communityName].readNextVars)
             else:
-                sys.stdout.write('Unsuppored PDU type %s from %s:%s\r\n' % (
+                log.msg('Unsuppored PDU type %s from %s:%s' % (
                     reqPDU.__class__.__name__, transportDomain,
                     transportAddress
                     ))
@@ -803,34 +894,31 @@ if v2cArch:
     
     # Configure socket server
    
-    sys.stdout.write('Listening at:\r\n')
- 
     transportDispatcher = AsynsockDispatcher()
     for idx in range(len(agentUDPv4Endpoints)):
         transportDispatcher.registerTransport(
-                udp.domainName + (idx,),
-                udp.UdpTransport().openServerMode(agentUDPv4Endpoints[idx])
+                udp.domainName + (idx,), agentUDPv4Endpoints[idx][0]
             )
-        sys.stdout.write('  UDP/IPv4 endpoint %s:%s, transport ID %s\r\n' % (agentUDPv4Endpoints[idx] + ('.'.join([str(x) for x in udp.domainName + (idx,)]),)))
+        log.msg('Listening at UDP/IPv4 endpoint %s, transport ID %s' % (agentUDPv4Endpoints[idx][1], '.'.join([str(x) for x in udp.domainName + (idx,)]),))
     for idx in range(len(agentUDPv6Endpoints)):
         transportDispatcher.registerTransport(
-                udp6.domainName + (idx,),
-                udp6.Udp6Transport().openServerMode(agentUDPv6Endpoints[idx])
+                udp6.domainName + (idx,), agentUDPv6Endpoints[idx][0]
             )
-        sys.stdout.write('  UDP/IPv6 endpoint %s:%s, transport ID %s\r\n' % (agentUDPv6Endpoints[idx] + ('.'.join([str(x) for x in udp6.domainName + (idx,)]),)))
+        log.msg('Listening at UDP/IPv6 endpoint %s, transport ID %s' % (agentUDPv6Endpoints[idx][1], '.'.join([str(x) for x in udp6.domainName + (idx,)]),))
     for idx in range(len(agentUNIXEndpoints)):
         transportDispatcher.registerTransport(
-                unix.domainName + (idx,),
-                unix.UnixTransport().openServerMode(agentUNIXEndpoints[idx])
+                unix.domainName + (idx,), agentUNIXEndpoints[idx][0]
             )
-        sys.stdout.write('  UNIX domain endpoint %s, transport ID %s\r\n' % (agentUNIXEndpoints[idx], '.'.join([str(x) for x in unix.domainName + (idx,)])))
+        log.msg('Listening at UNIX domain endpoint %s, transport ID %s' % (agentUNIXEndpoints[idx][1], '.'.join([str(x) for x in unix.domainName + (idx,)])))
     transportDispatcher.registerRecvCbFun(commandResponderCbFun)
 else:
-    sys.stdout.write('SNMPv3 credentials:\r\nUsername: %s\r\n' % v3User)
-    if authProtocols[v3AuthProto] != config.usmNoAuthProtocol:
-        sys.stdout.write('Authentication key: %s\r\nAuthentication protocol: %s\r\n' % (v3AuthKey, v3AuthProto))
-        if privProtocols[v3PrivProto] != config.usmNoPrivProtocol:
-            sys.stdout.write('Encryption (privacy) key: %s\r\nEncryption protocol: %s\r\n' % (v3PrivKey, v3PrivProto))
+    for v3User in v3Users:
+        log.msg('%s' % ('-'*66,))
+        log.msg('SNMPv3 USM SecurityName: %s' % v3User)
+        if authProtocols[v3AuthProtos[v3User]] != config.usmNoAuthProtocol:
+            log.msg('SNMPv3 USM authentication key: %s, authentication protocol: %s' % (v3AuthKeys[v3User], v3AuthProtos[v3User]))
+            if privProtocols[v3PrivProtos[v3User]] != config.usmNoPrivProtocol:
+                log.msg('SNMPv3 USM encryption (privacy) key: %s, encryption protocol: %s' % (v3PrivKeys[v3User], v3PrivProtos[v3User]))
 
     # Configure access to data index
 
@@ -843,29 +931,24 @@ else:
 
     # Configure socket server
 
-    sys.stdout.write('Listening at:\r\n')
-
     for idx in range(len(agentUDPv4Endpoints)):
         config.addSocketTransport(
             snmpEngine,
-            udp.domainName + (idx,),
-            udp.UdpTransport().openServerMode(agentUDPv4Endpoints[idx])
+            udp.domainName + (idx,), agentUDPv4Endpoints[idx][0]
         )
-        sys.stdout.write('  UDP/IPv4 endpoint %s:%s, transport ID %s\r\n' % (agentUDPv4Endpoints[idx] + ('.'.join([str(x) for x in udp.domainName + (idx,)]),)))
+        log.msg('Listening at UDP/IPv4 endpoint %s, transport ID %s' % (agentUDPv4Endpoints[idx][1], '.'.join([str(x) for x in udp.domainName + (idx,)]),))
     for idx in range(len(agentUDPv6Endpoints)):
         config.addSocketTransport(
             snmpEngine,
-            udp6.domainName + (idx,),
-            udp6.Udp6Transport().openServerMode(agentUDPv6Endpoints[idx])
+            udp6.domainName + (idx,), agentUDPv6Endpoints[idx][0]
         )
-        sys.stdout.write('  UDP/IPv6 endpoint %s:%s, transport ID %s\r\n' % (agentUDPv6Endpoints[idx] + ('.'.join([str(x) for x in udp6.domainName + (idx,)]),)))
+        log.msg('Listening at UDP/IPv6 endpoint %s, transport ID %s' % (agentUDPv6Endpoints[idx][1], '.'.join([str(x) for x in udp6.domainName + (idx,)]),))
     for idx in range(len(agentUNIXEndpoints)):
         config.addSocketTransport(
             snmpEngine,
-            unix.domainName + (idx,),
-            unix.UnixTransport().openServerMode(agentUNIXEndpoints[idx])
+            unix.domainName + (idx,), agentUNIXEndpoints[idx][0]
         )
-        sys.stdout.write('  UNIX domain endpoint %s, transport ID %s\r\n' % (agentUNIXEndpoints[idx], '.'.join([str(x) for x in unix.domainName + (idx,)])))
+        log.msg('Listening at UNIX domain endpoint %s, transport ID %s' % (agentUNIXEndpoints[idx][1], '.'.join([str(x) for x in unix.domainName + (idx,)])))
 
     # SNMP applications
 
@@ -887,25 +970,26 @@ exc_info = None
 try:
     transportDispatcher.runDispatcher()
 except KeyboardInterrupt:
-    sys.stdout.write('Process terminated\r\n')
+    log.msg('Shutting down process...')
 except Exception:
     exc_info = sys.exc_info()
 
 if variationModules:
-    sys.stdout.write('Shutting down variation modules:\r\n')
-    for name, body in variationModules.items():
-        sys.stdout.write('    %s...  ' % name)
+    log.msg('Shutting down variation modules:')
+    for name, contexts in variationModules.items():
+        body = contexts[0]
         try:
             body['shutdown'](not v2cArch and snmpEngine or None,
                              options=body['args'], mode='variation')
         except Exception:
-            sys.stdout.write('FAILED: %s\r\n' % sys.exc_info()[1])
+            log.msg('Module %s shutdown FAILED: %s' % (name, sys.exc_info()[1]))
         else:
-            sys.stdout.write('OK\r\n')
+            log.msg('Module %s shutdown OK' % name)
 
 transportDispatcher.closeDispatcher()
 
+log.msg('Process terminated')
+
 if exc_info:
-    e = exc_info[0](exc_info[1])
-    e.__traceback__ = exc_info[2]
-    raise e
+    for line in traceback.format_exception(*exc_info):
+        log.msg(line.replace('\n', ';'))
